/*
 * ambpack: packs, unpacks and lists AMB archives (Ancient Machine Book)
 *
 * http://ambook.sourceforge.net
 *
 * Copyright (C) 2020 Mateusz Viste
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <dirent.h>

#include <sys/stat.h>   /* mkdir */


#define PVER "20201201"
#define PDATE "2020"

#ifdef SLASHPATHSEP
#define PATHSEP "/"
#else
#define PATHSEP "\\"
#endif

static void amb_help(void) {
  puts("ambpack is an archiver that packs/unpacks AMB (Ancient Machine Book) files.");
  puts("");
  puts("usage: ambpack <action> <input_file|input_dir> [<output_file|output_dir>]");
  puts("");
  puts("actions:");
  puts("");
  puts("  c   creates an AMB file by packing the content of a directory");
  puts("  u   unpacks an AMB file to a directory of the same name");
  puts("  l   lists the content of an AMB file");
  puts("  t   tests an AMB file (checksums verification)");
  puts("");
  puts("ambpack ver " PVER " -- Copyright (C) " PDATE " Mateusz Viste");
}


static unsigned short bsd_sum(const void *buff, unsigned short flen) {
  unsigned short sum = 0;
  unsigned short i;
  for (i = 0; i < flen; i++) {
    unsigned short bsum_ror = (sum & 1) << 15;
    sum >>= 1;
    sum |= bsum_ror;
    sum += ((unsigned char *)buff)[i];
  }
  return(sum);
}


/* returns 0 on success, 1 on failure */
static int amb_create(const char *dirname, const char *outfile) {
  DIR *d = NULL;
  struct dirent *de = NULL;
  FILE *fd_amb = NULL;
  unsigned short i, fcount = 0;
  char *flist[65536];
  int ecode = 1; /* assume failure */

  /* try opening the output file (should fail, file is expected not to exist) */
  fd_amb = fopen(outfile, "rb");
  if (fd_amb != NULL) {
    printf("ERROR: file '%s' already exists", outfile);
    puts("");
    goto ERR_QUIT;
  }

  /* open directory */
  d = opendir(dirname);
  if (d == NULL) {
    printf("ERROR: failed to open directory '%s'", dirname);
    puts("");
    goto ERR_QUIT;
  }

  puts("Computing the list of files...");

  /* get the list of files */
  for (de = readdir(d); de != NULL; de = readdir(d)) {
    /* filter out unwanted things */
    if (de->d_name[0] == '.') continue; /* skip '.', '..' and hidden files */
    if (de->d_type == DT_DIR) {
      printf("ERROR: a subdirectory has been found (%s)", de->d_name);
      puts("");
      goto ERR_QUIT;
    }
    if (strlen(de->d_name) > 12) {
      printf("ERROR: filename too long (%s), max allowed is 12 characters", de->d_name);
      puts("");
      goto ERR_QUIT;
    }
    /* */
    if (fcount == 65535) {
      puts("ERROR: too many files in directory (max: 65535)");
      goto ERR_QUIT;
    }
    flist[fcount] = strdup(de->d_name);
    if (flist[fcount] == NULL) {
      puts("ERROR: out of memory");
      goto ERR_QUIT;
    }
    fcount++;
  }

  if (fcount == 0) {
    puts("ERROR: found no files to pack in directory");
    goto ERR_QUIT;
  }

  puts("Packing files...");

  fd_amb = fopen(outfile, "wb");
  if (fd_amb == NULL) {
    printf("ERROR: failed to create output file '%s'", outfile);
    puts("");
    goto ERR_QUIT;
  }

  /* write the AMB header */
  fprintf(fd_amb, "AMB1%c%c", fcount & 255, fcount >> 8);

  /* write a dummy (all-spaces) list of files */
  for (i = 0; i < fcount; i++) {
    if (fwrite("                    ", 1, 20, fd_amb) != 20) {
      puts("ERROR: write failure");
      goto ERR_QUIT;
    }
  }

  /* process each file and write it to the archive */
  for (i = 0; i < fcount; i++) {
    FILE *fd;
    uint16_t bsum = 0;
    uint16_t flen = 0;
    long offset;
    uint8_t buff[256];
    printf("  %s --> %s", flist[i], outfile);
    puts("");
    /* open file to be inserted into amb */
    snprintf((char *)buff, sizeof(buff), "%s" PATHSEP "%s", dirname, flist[i]);
    fd = fopen((char *)buff, "rb");
    if (fd == NULL) {
      printf("ERROR: failed to open file '%s'", buff);
      puts("");
      goto ERR_QUIT;
    }
    /* move amb file pointer to end of file and remember this position */
    fseek(fd_amb, 0, SEEK_END);
    offset = ftell(fd_amb);
    /* copy file to amb, computing file len at the same time and BSD sum */
    for (;;) {
      size_t t;
      size_t bytes = fread(buff, 1, sizeof(buff), fd);
      if (bytes == 0) break; /* EOF */
      /* len cannot be > 65535 */
      if (flen + bytes < flen) {
        puts("ERROR: file too big (max allowed size = 65535 bytes");
        fclose(fd);
        goto ERR_QUIT;
      }
      flen += bytes;
      /* */
      if (fwrite(buff, 1, bytes, fd_amb) != bytes) {
        puts("ERROR: write failure");
        fclose(fd);
        goto ERR_QUIT;
      }
      /* update bsum */
      for (t = 0; t < bytes; t++) {
        uint16_t bsum_ror = (bsum & 1) << 15;
        bsum >>= 1;
        bsum |= bsum_ror;
        bsum += buff[t];
      }
    }
    fclose(fd);
    /* update file header in amb file list */
    fseek(fd_amb, 4 + 2 + (i * 20), SEEK_SET);
    /* fname, zero-padded to 12 bytes */
    memset(buff, 0, 12);
    memcpy(buff, flist[i], strlen(flist[i]));
    /* offset */
    buff[12] = offset;
    buff[13] = offset >> 8;
    buff[14] = offset >> 16;
    buff[15] = offset >> 24;
    /* file len */
    buff[16] = flen;
    buff[17] = flen >> 8;
    /* bsum */
    buff[18] = bsum;
    buff[19] = bsum >> 8;
    fwrite(buff, 1, 20, fd_amb);
  }

  ecode = 0;

  ERR_QUIT:
  if (fd_amb != NULL) fclose(fd_amb);
  if (d != NULL) closedir(d);
  for (i = 0; i < fcount; i++) {
    free(flist[i]);
  }
  return(ecode);
}


static int amb_list(const char *fname) {
  FILE *fd = NULL;
  unsigned char buff[20];
  unsigned short fcount;
  fd = fopen(fname, "rb");
  if (fd == NULL) {
    printf("ERROR: failed to open file '%s'", fname);
    puts("");
    goto ERR_QUIT;
  }
  /* read header */
  if ((fread(buff, 1, 6, fd) != 6) || (memcmp(buff, "AMB1", 4) != 0)) {
    puts("ERROR: not an AMB archive");
    goto ERR_QUIT;
  }
  fcount = buff[5];
  fcount <<= 8;
  fcount = buff[4];
  /* read file entries */
  while (fcount--) {
    if (fread(buff, 1, 20, fd) != 20) {
      puts("ERROR: archive is corrupted");
      goto ERR_QUIT;
    }
    /* print file info (filename string limited to 12 characters) */
    printf("%.12s  (%u bytes)", (char *)buff, buff[16] | (buff[17] << 8));
    puts("");
  }

  fclose(fd);
  return(0);

  ERR_QUIT:
  if (fd != NULL) fclose(fd);
  return(1);
}


static int amb_unpack(const char *fname, const char *dirname) {
  FILE *fd = NULL;
  int ecode = 1;
  unsigned char buff[65535];
  unsigned short i, fcount;

  fd = fopen(fname, "rb");
  if (fd == NULL) {
    printf("ERROR: failed to open file '%s'", fname);
    puts("");
    goto ERR_QUIT;
  }

  /* read header */
  if ((fread(buff, 1, 6, fd) != 6) || (memcmp(buff, "AMB1", 4) != 0)) {
    puts("ERROR: not an AMB archive");
    goto ERR_QUIT;
  }
  fcount = buff[5];
  fcount <<= 8;
  fcount = buff[4];

  /* create the target directory (unless I run in NULL mode) */
  if ((dirname != NULL) && (mkdir(dirname, S_IRWXU) != 0)) {
    printf("ERROR: failed to create directory '%s'", dirname);
    puts("");
    goto ERR_QUIT;
  }

  /* iterate over file entries */
  for (i = 0; i < fcount; i++) {
    FILE *fdout;
    long offset;
    unsigned short flen, bsum, mybsum;
    unsigned char hdr[20];
    fseek(fd, 4 + 2 + (20 * i), SEEK_SET);
    if (fread(hdr, 1, 20, fd) != 20) {
      puts("ERROR: archive is corrupted");
      goto ERR_QUIT;
    }
    offset = hdr[15];
    offset <<= 8;
    offset |= hdr[14];
    offset <<= 8;
    offset |= hdr[13];
    offset <<= 8;
    offset |= hdr[12];
    flen = hdr[17];
    flen <<= 8;
    flen |= hdr[16];
    bsum = hdr[19];
    bsum <<= 8;
    bsum |= hdr[18];
    /* compute dest file name */
    if (dirname != NULL) {
      snprintf((char *)buff, sizeof(buff), "%s" PATHSEP "%.12s", dirname, (char *)hdr);
    } else {
      snprintf((char *)buff, sizeof(buff), "%.12s", (char *)hdr);
    }
    /* output a log */
    printf(" --> %.12s (%u bytes, bsum=0x%04X)", (char *)hdr, flen, bsum);
    puts("");
    /* create dst file (unless null mode) */
    if (dirname != NULL) {
      fdout = fopen((char *)buff, "wb");
      if (fdout == NULL) {
        puts("ERROR: failed to create output file");
        goto ERR_QUIT;
      }
    }
    /* read the file from archive and write it to dest (unless NULL mode) */
    if (fread(buff, 1, flen, fd) != flen) {
      puts("ERROR: read failure from archive");
      fclose(fdout);
      goto ERR_QUIT;
    }
    if (dirname != NULL) {
      if (fwrite(buff, 1, flen, fdout) != flen) {
        puts("ERROR: write failure to dest file");
        fclose(fdout);
        goto ERR_QUIT;
      }
      fclose(fdout);
    }
    /* compute and match the BSD sum */
    mybsum = bsd_sum(buff, flen);
    if (mybsum != bsum) {
      printf("ERROR: checksum mistmatch (expected=%04X != computed=%04X), the archive file appears to be corrupted!", bsum, mybsum);
      puts("");
      goto ERR_QUIT;
    }
  }

  ecode = 0;

  ERR_QUIT:
  if (fd != NULL) fclose(fd);
  return(ecode);
}


int main(int argc, char **argv) {
  int ecode;

  if (argc < 2) {
    amb_help();
    return(1);
  }

  /* recognize the action */
  switch (argv[1][0]) {
    case 'c':
    case 'C':
      if (argc != 4) {
        amb_help();
        ecode = 1;
      } else {
        ecode = amb_create(argv[2], argv[3]);
      }
      break;
    case 'u':
    case 'U':
      if (argc != 4) {
        amb_help();
        ecode = 1;
      } else {
        ecode = amb_unpack(argv[2], argv[3]);
      }
      break;
    case 'l':
    case 'L':
      if (argc != 3) {
        amb_help();
        ecode = 1;
      } else {
        ecode = amb_list(argv[2]);
      }
      break;
    case 't':
    case 'T':
      if (argc != 3) {
        amb_help();
        ecode = 1;
      } else {
        ecode = amb_unpack(argv[2], NULL);
      }
      break;
    default:
      amb_help();
      ecode = 1;
      break;
  }

  if (ecode == 0) puts("DONE");
  return(ecode);
}
